AssetBrowser = class(Turbine.UI.Lotro.Window);

function AssetBrowser:Constructor(initialAssetID)
    Turbine.UI.Lotro.Window.Constructor(self);

    self:SetText("Asset Browser");
    self:SetResizable(true);
    self:SetMinimumSize(571, 244);
    self:SetVisible(true);
    self:SetWantsUpdates(true);

    self.precedingAssetIDIndices, self.followingAssetIDIndices = {}, {};
    self.precedingCells, self.followingCells = {}, {};
    self.gridSizeX, self.gridSizeY = 1, 1;
    self.filterChangeTime = Turbine.Engine.GetGameTime();
    self.imgBackColor = nil;
    self.imgBackColorBlendMode = Turbine.UI.BlendMode.Overlay;
    self.imgBlendMode = Turbine.UI.BlendMode.Normal;

    self:CreateControls();

    if (initialAssetID) then
        local idx = assetIDIndex[initialAssetID];
        if (not idx) then
            Turbine.Shell.WriteLine("Invalid asset ID: " .. tostring(initialAssetID));
            self:Close();
            return;
        else
            self:SetSize(self:GetMinimumSize());
            self.gridCheckBox:SetChecked(false);
            self.gridXScrollBar:SetValue(7);
            self.gridYScrollBar:SetValue(1);
            self:ResetFilter();
        end
    else
        -- Show some interesting images in initial view
        self:SetSize(1028, 720);
        self:ResetFilter();
        self.wantOld:SetChecked(false);
        self.wantUd20x:SetChecked(false);
        self.wantUd21x:SetChecked(false);
        self.wantUd22x:SetChecked(false);
        self.wantUd23x:SetChecked(false);
        self.wantUd24x:SetChecked(true);
--        self.widthMin:SetText(200);
--        self.widthMax:SetText(200);
--        self.heightMin:SetText(200);
--        self.heightMax:SetText(200);
        self.gridCheckBox:SetChecked(true);
        self.gridXScrollBar:SetValue(3);
        self.gridYScrollBar:SetValue(1);
        initialAssetID = 0x411DA6CE;
    end
    self.assetIDField:SetText("0x" .. string.format("%8.8X", initialAssetID));
    self.assetIDScrollBar:SetValue(assetIDIndex[initialAssetID]);
    self.gridCheckBox:CheckedChanged();
end

function AssetBrowser:ResetFilter()
    self.wantOld:SetChecked(true);
    self.wantUd20x:SetChecked(true);
    self.wantUd21x:SetChecked(true);
    self.wantUd22x:SetChecked(true);
    self.wantUd23x:SetChecked(true);
    self.wantUd24x:SetChecked(true);
    self.widthMin:SetText(minAssetWidth);
    self.widthMax:SetText(maxAssetWidth);
    self.heightMin:SetText(minAssetHeight);
    self.heightMax:SetText(maxAssetHeight);
    self.aspectMin:SetText(minAssetAspectRatio[2]);
    self.aspectMax:SetText(maxAssetAspectRatio[2]);
end

function AssetBrowser:CreateControls()
    local function AddLabel(parent, top, left, text, width)
        local label = Turbine.UI.Label();
        label:SetParent(parent);
        label:SetPosition(left, top + 1);
        label:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight);
        label:SetSize(width, 18);
        label:SetText(text);
        return label, left + width;
    end

    local function AddCheckBox(parent, top, left, text, width)
        local checkBox = Turbine.UI.Lotro.CheckBox();
        checkBox:SetParent(parent);
        checkBox:SetPosition(left, top);
        checkBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        checkBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
        checkBox:SetSize(width, 20);
        checkBox:SetText(text);
        checkBox.GetValue = function(ctl)
            return ctl:IsChecked();
        end
        return checkBox, left + width;
    end
    
    local function AddTextBox(parent, top, left, width)
        local textBox = Turbine.UI.Lotro.TextBox();
        textBox:SetParent(parent);
        textBox:SetPosition(left, top + 1);
        textBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        textBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
        textBox:SetSize(width, 18);
        textBox:SetMultiline(false);
        textBox.GetValue = function(ctl)
            return tonumber(ctl:GetText());
        end
        return textBox, left + width;
    end
    
    local function AddButton(parent, top, left, text, width, font)
        local button = Turbine.UI.Lotro.Button();
        button:SetParent(parent);
        button:SetPosition(left, top);
        button:SetWidth(width);
        if (font) then
            button:SetFont(font);
        end
        button:SetText(text);
        return button, left + width;
    end
    
    local function AddScrollBar(parent, minValue, maxValue)
        local scrollBar = Turbine.UI.Lotro.ScrollBar();
        scrollBar:SetParent(parent);
        scrollBar:SetOrientation(Turbine.UI.Orientation.Horizontal);
        scrollBar:SetMinimum(minValue);
        scrollBar:SetMaximum(maxValue);
        scrollBar:SetSize(100, 10);
        return scrollBar;
    end
    
    local left, top;
    local label, checkBox, textBox, button, scrollBar;
    
    local container = Turbine.UI.Control();
    container:SetParent(self);
    container:SetTop(20);
    container:SetBackColor(Turbine.UI.Color(0.95, 0.15, 0.15, 0.15));
    self.filterControl = container;

    local xBorder, yBorder = 6, 3;
    left, top = xBorder, yBorder;
    label, left = AddLabel(container, top, left, "Date: ", 48);
    
    self.wantOld, left = AddCheckBox(container, top, left, "Old", 45);
 
    self.wantUd20x, left = AddCheckBox(container, top, left, "Update 20", 83);

    self.wantUd21x, left = AddCheckBox(container, top, left, "21", 40);

    self.wantUd22x, left = AddCheckBox(container, top, left, "22", 40);

    self.wantUd23x, left = AddCheckBox(container, top, left, "23", 40);

    self.wantUd24x, left = AddCheckBox(container, top, left, "24", 40);

    top, left = yBorder + 20, xBorder;
    _, left = AddLabel(container, top, left, "Width: ", 48);

    self.widthMin, left = AddTextBox(container, top, left, 35);

    _, left = AddLabel(container, top, left, "-", 6);

    self.widthMax, left = AddTextBox(container, top, left, 35);
    
    _, left = AddLabel(container, top, left, "Height: ", 65);
    
    self.heightMin, left = AddTextBox(container, top, left, 35);

    _, left = AddLabel(container, top, left, "-", 6);

    self.heightMax, left = AddTextBox(container, top, left, 35);

    _, left = AddLabel(container, top, left, "W/H Ratio: ", 90);

    textBox, left = AddTextBox(container, top, left, 45);
    textBox.GetValue = function(ctl)
        local func = loadstring("return " .. ctl:GetText());
        return (func and func() or nil);
    end
    self.aspectMin = textBox;

    _, left = AddLabel(container, top, left, "-", 6);

    textBox, left = AddTextBox(container, top, left, 45);
    textBox.GetValue = function(ctl)
        local func = loadstring("return " .. ctl:GetText());
        return (func and func() or nil);
    end
    self.aspectMax = textBox;

    button, left = AddButton(container, yBorder, left - 95, "Reset Filter", 95);
    button:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
    button.Click = function()
        self:ResetFilter();
    end

    container:SetSize(left + xBorder, top + 20 + yBorder);

    container = Turbine.UI.Control();
    container:SetParent(self);
    self.assetIDControl = container;
    left, top = 0, 0;
    
    button, left = AddButton(container, top, left, "◄", 40, Turbine.UI.Lotro.Font.Arial12);
    button.Click = function(ctl)
        self:SelectAsset(ctl.prevIdx);
    end
    self.prevButton = button;

    _, left = AddLabel(container, top, left, "ID: ", 30);

    textBox, left = AddTextBox(container, top, left, 95);
    textBox.SetText = function(ctl, value)
        Turbine.UI.Lotro.TextBox.SetText(ctl, value);
        ctl.lastKnownValue = value;
    end
    textBox.GetValue = function(ctl)
        local id = tonumber(ctl:GetText());
        if (id) then
            return assetIDIndex[id];
        end
    end
    self.assetIDField = textBox;

    left = left + label:GetWidth();
    button, left = AddButton(container, top, left, "►", 40, Turbine.UI.Lotro.Font.Arial12);
    button.Click = function(ctl)
        self:SelectAsset(ctl.nextIdx);
    end
    self.nextButton = button;
    
    container:SetSize(left, top + 20);

    local scrollBar = AddScrollBar(self, 1, numAssets);
    scrollBar.SetValue = function(ctl, value)
        Turbine.UI.Lotro.ScrollBar.SetValue(ctl, value);
        ctl.lastKnownValue = ctl:GetValue();
    end
    self.assetIDScrollBar = scrollBar;

    self.imageBorder = Turbine.UI.Lotro.TextBox();
    self.imageBorder:SetParent(self);
    self.backdrop = Turbine.UI.Control();
    self.backdrop:SetParent(self.imageBorder);
    self.backdrop:SetBackground("Thurallor/AssetBrowser/backdrop.tga");

    self.imageOverlay = Turbine.UI.Window();
    self.imageOverlay:SetVisible(true);
    self.imageOverlay:SetZOrder(self:GetZOrder() + 1);
    self.imageOverlay:SetWantsUpdates(true);
    self.imageOverlay.Update = function(ctl)
        local mouseX, mouseY = ctl:GetMousePosition();
        local w, h = ctl:GetSize();
        if ((mouseX > 0) and (mouseY > 0) and (mouseX < w) and (mouseY < h)) then
            self.backdrop:SetPosition((mouseX / 20) % 64 - 64, (mouseY / 20) % 64 - 64);
        end
    end
    
    self.centerCell = Turbine.UI.Control();
    self.centerCell.outline = Turbine.UI.Control();
    self.centerCell.outline:SetParent(self.imageOverlay);
    self.centerCell.interior = Turbine.UI.Control();
    self.centerCell.interior:SetParent(self.centerCell.outline);
    self.centerCell.interior:SetPosition(1, 1);
    self.centerCell:SetParent(self.imageOverlay);
    self.centerCell:SetVisible(true);
    self.centerCell.nativeWidth = 0;
    self.centerCell.nativeHeight = 0;
    self.centerCell.MouseClick = function(ctl, args)
        -- Left-click: Recenter grid on this cell.
        if (args.Button == Turbine.UI.MouseButton.Left) then
            self:SelectAsset(ctl.assetIDIndex);

        
        -- Right-click: Show context menu.
        elseif (args.Button == Turbine.UI.MouseButton.Right) then
            local contextMenu = Turbine.UI.ContextMenu();
            local idStr = "ID: " .. string.format("0x%8.8X", ctl.assetID);
            local sizeStr = "Size: " .. tostring(ctl.nativeWidth) .. " × " .. tostring(ctl.nativeHeight);
            local version = assetVersion[assetIDIndex[ctl.assetID]];
            local versionStr = "Appeared in: Update " .. ((version > 0) and tostring(version / 10) or "12.0 or before");
            contextMenu:GetItems():Add(Turbine.UI.MenuItem(idStr, false, false));
            contextMenu:GetItems():Add(Turbine.UI.MenuItem(sizeStr, false, false));
            contextMenu:GetItems():Add(Turbine.UI.MenuItem(versionStr, false, false));
            local clickableItem = Turbine.UI.MenuItem("Copy to chat", true, false);
            clickableItem.Click = function()
                Turbine.Shell.WriteLine(idStr .. "; " .. sizeStr .. "; " .. versionStr);
            end
            contextMenu:GetItems():Add(clickableItem);
            contextMenu:ShowMenu();
        end
    end

    self.progressBar = Thurallor.UI.ProgressBar();
    self.progressBar:SetParent(self);
    self.progressBar:SetHeight(2);
    self.progressBar:SetProgress(0);
    local function updateProgressBar()
        local progress = self.leftProgressBar:GetProgress() + self.rightProgressBar:GetProgress();
        if ((progress > 0) and (progress < 2)) then
            self.progressBar:SetParent(self);
            self.progressBar:SetProgress(progress / 2);
        else
            self.progressBar:SetParent(nil);
        end
    end
    self.leftProgressBar = Thurallor.UI.ProgressBar();
    self.leftProgressBar:SetProgress(0);
    self.leftProgressBar.ProgressChanged = function(ctl)
        updateProgressBar();
    end
    self.rightProgressBar = Thurallor.UI.ProgressBar();
    self.rightProgressBar:SetProgress(0);
    self.rightProgressBar.ProgressChanged = function(ctl)
        updateProgressBar();
    end
    
    checkBox = AddCheckBox(self, 0, 0, "Tiled", 60);
    checkBox.CheckedChanged = function(ctl)
        if (ctl:IsChecked()) then
            self.scaledCheckBox:SetChecked(false);
        end
        self:UpdateImageDisplay();
    end
    self.tiledCheckBox = checkBox;

    checkBox = AddCheckBox(self, 0, 0, "Scaled", 69);
    checkBox.CheckedChanged = function(ctl)
        if (ctl:IsChecked()) then
            self.tiledCheckBox:SetChecked(false);
        end
        self:UpdateImageDisplay();
    end
    self.scaledCheckBox = checkBox;

    label = AddLabel(self, 0, 0, "Render...", 60);
    label:SetMarkupEnabled(true);
    label.MouseEnter = function ()
        label:SetText("<u>Render...</u>");
    end
    label.MouseLeave = function()
        label:SetText("Render...");
    end
    label.MouseClick = function()
        self:ShowRenderMenu();
    end
    self.renderLabel = label;

    checkBox = AddCheckBox(self, 0, 0, "", 16)
    checkBox.CheckedChanged = function(ctl)
        self:UpdateImageDisplay();
    end
    self.renderCheckBox = checkBox;

    button = AddButton(self, 0, 0, "◄◄", 40, Turbine.UI.Lotro.Font.Arial12);
    button.Click = function(ctl)
        self:SelectAsset(ctl.centerIdx);
    end
    self.prevPageButton = button;

    button = AddButton(self, 0, 0, "►►", 40, Turbine.UI.Lotro.Font.Arial12);
    button.Click = function(ctl)
        self:SelectAsset(ctl.centerIdx);
    end
    self.nextPageButton = button;

    checkBox = AddCheckBox(self, 0, 0, "Grid", 55);
    checkBox.CheckedChanged = function(ctl)
        local checked = ctl:IsChecked();
        ctl:SetText(checked and "Grid: " or "Grid");
        self.gridXScrollBar:SetParent(checked and self or nil);
        self.gridYScrollBar:SetParent(checked and self or nil);
        self:ResizeGrid();
    end
    self.gridCheckBox = checkBox;

    self.sizeLabel = AddLabel(self, 0, 0, "", 100);
    
    scrollBar = AddScrollBar(self, 2, 33);
    scrollBar.ValueChanged = function()
        self:ResizeGrid();
    end
    self.gridXScrollBar = scrollBar;

    scrollBar = AddScrollBar(self, 2, 33);
    scrollBar.ValueChanged = function()
        self:ResizeGrid();
    end
    self.gridYScrollBar = scrollBar;
end

function AssetBrowser:Closing()
    self.imageOverlay:SetVisible(false);
end

function AssetBrowser:SizeChanged()
    local W, H = self:GetSize();
    local top, left, width, height = 35;
    
    width, height = self.filterControl:GetSize();
    left = math.floor(0.5 + (W / 2) - (width / 2));
    self.filterControl:SetPosition(left, top);
    top = top + height + 2;

    local assetIDControlHeight = 0;
    if (self.assetIDControl:IsVisible()) then
        width, height = self.assetIDControl:GetSize();
        left = math.floor(0.5 + (W / 2) - (width / 2));
        self.assetIDControl:SetPosition(left, top);
        top = top + height + 1;
        assetIDControlHeight = height;
    else
        top = top + 2;
    end
    
    left, width, height = 10, W - 20, 10;
    self.assetIDScrollBar:SetPosition(left, top);
    self.assetIDScrollBar:SetSize(width, height);
    top = top + height;
    
    left, width, height = 10, W - 20, H - (98 + self.filterControl:GetHeight() + assetIDControlHeight);
    self.imageBorder:SetPosition(left, top);
    self:PositionChanged();
    self.imageBorder:SetSize(width, height);
    self.backdrop:SetSize(width + 64, height + 64);
    self.imageOverlay:SetSize(width, height);
    self.progressBar:SetWidth(width);
    self.progressBar:SetPosition(left, top + height);
    top = top + height + 10;
    
    width = self.sizeLabel:GetWidth();
    left = math.floor(0.5 + (W / 2) - (width / 2));
    self.sizeLabel:SetPosition(left, top);
    width = self.prevPageButton:GetWidth() + 10 + self.nextPageButton:GetWidth();
    left = math.floor(0.5 + (W / 2) - (width / 2));
    self.prevPageButton:SetPosition(left, top - 1);
    left = left + self.prevPageButton:GetWidth() + 10;
    self.nextPageButton:SetPosition(left, top - 1);
    left = 30;
    self.tiledCheckBox:SetPosition(left, top);
    left = left + self.tiledCheckBox:GetWidth();
    self.scaledCheckBox:SetPosition(left, top);
    left = left + self.scaledCheckBox:GetWidth();
    self.renderCheckBox:SetPosition(left, top);
    left = left + self.renderCheckBox:GetWidth() - 6;
    self.renderLabel:SetPosition(left, top + 1);
    local right = W - 30;
    left = right - self.gridXScrollBar:GetWidth() - self.gridCheckBox:GetWidth();
    self.gridCheckBox:SetPosition(left, top);
    left = left + self.gridCheckBox:GetWidth();
    if (self.gridCheckBox:IsChecked()) then
        self.gridXScrollBar:SetPosition(left, top);
        top = top + 10;
        self.gridYScrollBar:SetPosition(left, top);
    end
    
    local borderWidth, borderHeight = self.imageBorder:GetSize();
    self.cellSizeX = math.floor(borderWidth / self.gridSizeX);
    self.cellSizeY = math.floor(borderHeight / self.gridSizeY);

    self:UpdateNeeded();
end

function AssetBrowser:PositionChanged()
    self.imageOverlay:SetPosition(self:PointToScreen(self.imageBorder:GetPosition()));
end

function AssetBrowser:LayoutImage(image, L, T, W, H)
    -- Add margins:
    L = L + 1; T = T + 1; W = W - 1; H = H - 1;
    if (self.renderCheckBox:IsChecked()) then
        image:SetBackColor(self.imgBackColor);
        image:SetBackColorBlendMode(self.imgBackColorBlendMode);
        image:SetBlendMode(self.imgBlendMode);
    else
        image:SetBackColor(nil);
        image:SetBackColorBlendMode(Turbine.UI.BlendMode.Overlay);
        image:SetBlendMode(Turbine.UI.BlendMode.Normal);
    end
    image:SetVisible(false); -- delete image residue from screen
    image:SetVisible(true);
    image:SetStretchMode(2); -- resize to image's native size
    image.nativeWidth, image.nativeHeight = image:GetWidth(), image:GetHeight();
    local left, top, width, height;
    if (self.tiledCheckBox:IsChecked()) then
        left, top = L, T;
        width, height = W, H;
        image:SetStretchMode(0);
    elseif (self.scaledCheckBox:IsChecked()) then
        image:SetStretchMode(1);
        local horizStretch = W / image.nativeWidth;
        local vertStretch = H / image.nativeHeight;
        local scale = math.min(horizStretch, vertStretch);
        width = image.nativeWidth * scale;
        height = image.nativeHeight * scale;
        left = L + math.floor(0.5 + (W / 2) - (width / 2));
        top = T + math.floor(0.5 + (H / 2) - (height / 2));
    else -- centered
        width, height = image.nativeWidth, image.nativeHeight;
        image:SetStretchMode(0);
        left = L + math.floor(0.5 + (W / 2) - (width / 2));
        top = T + math.floor(0.5 + (H / 2) - (height / 2));
        if (left < L) then
            left = L;
            width = W;
        end
        if (top < T + 1) then
            top = T;
            height = H;
        end
    end
    if (self.renderCheckBox:IsChecked() and self.imgOutline) then
        image.outline:SetPosition(left - 1, top - 1);
        image.outline:SetSize(width + 2, height + 2);
        image.outline:SetBackColor(Turbine.UI.Color.White);
        image.interior:SetSize(width, height);
        image.interior:SetBackColor(nil);
        image.interior:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
    else
        image.outline:SetBackColor(nil);
        image.interior:SetBackColor(nil);
    end
    image:SetPosition(left, top);
    image:SetSize(width, height);
end

function AssetBrowser:ResetGridDrawing()
    local gridItems = self.gridSizeX * self.gridSizeY;
    local precedingItems = (gridItems - 1) / 2;
    for i = 1, precedingItems, 1 do
        if (self.precedingCells[i] ~= nil) then
            self.precedingCells[i]:SetVisible(false);
        end
        if (self.followingCells[i] ~= nil) then
            self.followingCells[i]:SetVisible(false);
        end
    end
    self.drawingStep = 1;
end

function AssetBrowser:ResizeGrid()
    if (self.gridCheckBox:IsChecked()) then
        self.gridSizeX = self.gridXScrollBar:GetValue() * 2 - 1;
        self.gridSizeY = self.gridYScrollBar:GetValue() * 2 - 1;
    else
        self.gridSizeX = 1;
        self.gridSizeY = 1;
    end
    local gridItems = self.gridSizeX * self.gridSizeY;
    local showingGrid = (gridItems > 1);
    self.assetIDControl:SetVisible(not showingGrid);
    self.sizeLabel:SetVisible(not showingGrid);
    self.prevPageButton:SetVisible(showingGrid);
    self.nextPageButton:SetVisible(showingGrid);
    local precedingItems = (gridItems - 1) / 2;
    for i = 1, precedingItems, 1 do
        if (self.precedingCells[i] == nil) then
            self.precedingCells[i] = Turbine.UI.Control();
            self.precedingCells[i].outline = Turbine.UI.Control();
            self.precedingCells[i].interior = Turbine.UI.Control();
            self.precedingCells[i].interior:SetParent(self.precedingCells[i].outline);
            self.precedingCells[i].MouseClick = self.centerCell.MouseClick;
            self.followingCells[i] = Turbine.UI.Control();
            self.followingCells[i].outline = Turbine.UI.Control();
            self.followingCells[i].interior = Turbine.UI.Control();
            self.followingCells[i].interior:SetParent(self.followingCells[i].outline);
            self.followingCells[i].MouseClick = self.centerCell.MouseClick;
        end
        self.precedingCells[i].xIndex, self.precedingCells[i].yIndex = self:GetCellIndices(-i);
        self.precedingCells[i].outline:SetParent(self.imageOverlay);
        self.precedingCells[i].interior:SetPosition(1, 1);
        self.precedingCells[i]:SetParent(self.imageOverlay);
        self.followingCells[i].xIndex, self.followingCells[i].yIndex = self:GetCellIndices(i);
        self.followingCells[i].outline:SetParent(self.imageOverlay);
        self.followingCells[i].interior:SetPosition(1, 1);
        self.followingCells[i]:SetParent(self.imageOverlay);
    end
    
    -- If the grid has gotten smaller, hide the unneeded cells.
    for i = precedingItems + 1, #self.precedingCells, 1 do
        self.precedingCells[i].outline:SetParent(nil);
        self.precedingCells[i]:SetParent(nil);
        self.precedingCells[i] = nil;
        self.followingCells[i].outline:SetParent(nil);
        self.followingCells[i]:SetParent(nil);
        self.followingCells[i] = nil;
    end
    
    self:SizeChanged(); -- also calls UpdateImageDisplay()
end

function AssetBrowser:GetCellPosition(x, y)
    return self.cellSizeX * (x - 1), self.cellSizeY * (y - 1);
end

function AssetBrowser:GetCenterCellIndices()
    local x = (self.gridSizeX - 1) / 2 + 1;
    local y = (self.gridSizeY - 1) / 2 + 1;
    return x, y;
end

function AssetBrowser:GetCellIndices(num)
    local centerX, centerY = self:GetCenterCellIndices();
    local cellX, cellY;
    if (num > 0) then
        local x = centerX + num - 1;
        cellX = math.fmod(x, self.gridSizeX) + 1;
        cellY = centerY + math.floor(x / self.gridSizeX);
    else
        cellX, cellY = self:GetCellIndices(-num);
        cellX = self.gridSizeX - cellX + 1;
        cellY = self.gridSizeY - cellY + 1;
    end
--Turbine.Shell.WriteLine("GetCellIndices(" .. num .. ") = " .. cellX .. ", " .. cellY);
    return cellX, cellY;
end

function AssetBrowser:InvalidateImage()
    -- Filter has changed.  Hide the image display area, which is probably no longer valid.
    self.leftProgressBar:Cancel();
    self.rightProgressBar:Cancel();
    self.imageOverlay:SetVisible(false);
end

function AssetBrowser:UpdateNeeded()

    self:InvalidateImage();

    -- After we've stopped editing for 0.5s, redisplay the image area.
    self.updateNeeded = true;
    self.filterChangeTime = Turbine.Engine.GetGameTime();
end

function AssetBrowser:UpdateImageDisplay()
    local gridX, gridY = self:GetCenterCellIndices();
    local left, top = self:GetCellPosition(gridX, gridY);
    self:LayoutImage(self.centerCell, left, top, self.cellSizeX, self.cellSizeY);
    self.sizeLabel:SetText(tostring(self.centerCell.nativeWidth) .. " × " .. tostring(self.centerCell.nativeHeight));
    local gridItems = self.gridSizeX * self.gridSizeY;
    if (gridItems > 1) then
        self.precedingAssetIDIndices[0] = self.centerCell.assetIDIndex;
        self.followingAssetIDIndices[0] = self.centerCell.assetIDIndex;
        self:ResetGridDrawing();
        self.prevPageButton:SetEnabled(false);
        self.nextPageButton:SetEnabled(false);
        self:UpdateGridCenter();
    else
        self.prevButton:SetEnabled(false);
        self.prevButton.prevIdx = nil;
        self.nextButton:SetEnabled(false);
        self.nextButton.nextIdx = nil;
        self:_SearchLeft(self.assetIDIndex, 1, false, function(matches)
            if (matches[1]) then
                self.prevButton:SetEnabled(true);
                self.prevButton.prevIdx = matches[1];
            end
        end);
        self:_SearchRight(self.assetIDIndex, 1, false, function(matches)
            if (matches[1]) then
                self.nextButton:SetEnabled(true);
                self.nextButton.nextIdx = matches[1];
            end
        end);
    end
end

function AssetBrowser:UpdateGridCenter()
    local gridItems = self.gridSizeX * self.gridSizeY;
    local sideItems = (gridItems - 1) / 2;
    local wantedMatches =  sideItems * 2 + 1; -- keep searching to find center of adjacent page
    self:_SearchLeft(self.assetIDIndex, wantedMatches, false, function(matches, done)
        local m = #matches;
        if (m > 0) then
            if (m <= sideItems) then
                local idx = matches[m];
                self.precedingAssetIDIndices[m] = idx;
                local cell = self.precedingCells[m];
                cell.assetID, cell.assetIDIndex = assetID[idx], idx;
                pcall(cell.SetBackground, cell, cell.assetID);
                local left, top = self:GetCellPosition(self:GetCellIndices(-m));
                self:LayoutImage(cell, left, top, self.cellSizeX, self.cellSizeY);
            elseif (done and m > sideItems) then
                self.prevPageButton:SetEnabled(true);
                self.prevPageButton.centerIdx = matches[m];
            end
        end
    end);
    self:_SearchRight(self.assetIDIndex, wantedMatches, false, function(matches, done)
        local m = #matches;
        if (m > 0) then
            if (m <= sideItems) then
                local idx = matches[m];
                self.followingAssetIDIndices[m] = idx;
                local cell = self.followingCells[m];
                cell.assetID, cell.assetIDIndex = assetID[idx], idx;
                pcall(cell.SetBackground, cell, cell.assetID);
                local left, top = self:GetCellPosition(self:GetCellIndices(m));
                self:LayoutImage(cell, left, top, self.cellSizeX, self.cellSizeY);
            elseif (done and m > sideItems) then
                self.nextPageButton:SetEnabled(true);
                self.nextPageButton.centerIdx = matches[m];
            end
        end
    end);
end

function AssetBrowser:MatchesFilter(id)
    local idx = assetIDIndex[id];
    local version = assetVersion[idx] / 10;
    local params = self.filterParams;
    if (version < 20.0) then
        if (not params.wantOld) then
            return false;
        end
    elseif ((version >= 20.0) and (version < 21.0)) then
        if (not params.wantUd20x) then
            return false;
        end
    elseif ((version >= 21.0) and (version < 22.0)) then
        if (not params.wantUd21x) then
            return false;
        end
    elseif ((version >= 22.0) and (version < 23.0)) then
        if (not params.wantUd22x) then
            return false;
        end
    elseif ((version >= 23.0) and (version < 24.0)) then
        if (not params.wantUd23x) then
            return false;
        end
    elseif (version >= 24.0) then
        if (not params.wantUd24x) then
            return false;
        end
    end
    local width, height = unpack(assetSize[idx]);
    local aspectRatio = width / height;
    local params = self.filterParams;
    if ((width < params.widthMin) or (width > params.widthMax) or (height < params.heightMin) or (height > params.heightMax) or (aspectRatio < params.aspectMin) or (aspectRatio > params.aspectMax)) then
        return false;
    end
    return true;
end

function AssetBrowser:SelectAsset(idx, updateField, updateScrollBar)
    self.imageOverlay:SetVisible(idx ~= nil);
    if (not idx) then
        return;
    end
    local id = assetID[idx];
    self.assetID, self.assetIDIndex = id, idx;
    self.centerCell.assetID, self.centerCell.assetIDIndex = id, idx;
    pcall(self.centerCell.SetBackground, self.centerCell, id);
    self.precedingAssetIDIndices = {};
    self.followingAssetIDIndices = {};
    self:UpdateImageDisplay();
    self.assetIDField:SetText(string.format("0x%8.8X", id));
    self.assetIDScrollBar:SetValue(idx);
end

-- Searches to the right to finds one or more assets ('matchesWanted') that match the filter,
-- starting at the specified index ('idx') + 1.
-- The 'wrap' arg indicates whether wrapping is allowed.
-- Each time a match is found, and at the end of the search regardless of whether any
-- matches were found, the callback function ('callbackFunc') is called.  Its arguments are
--    1. a list of the matching asset indexes, which may be empty
--    2. a boolean value indicating whether the operation is complete
function AssetBrowser:_SearchRight(idx, matchesWanted, wrap, callbackFunc)
--idxStr = "nil";
--if (idx) then
--    idxStr = idx .. "=>" .. string.format("0x%8.8X", assetID[idx]);
--end
--Puts("SearchRight(" .. idxStr .. ", " .. tostring(matchesWanted) .. ", " .. tostring(wrap) .. ", func)");
    if ((idx == nil) or ((not wrap) and (idx == numAssets))) then
        -- Nothing to do.
        callbackFunc({}, true);
        return;
    end
    self.searchRight = {
        idx = idx + 1;
        matchesWanted = matchesWanted;
        stopIdx = 1;
        matches = {};
        callbackFunc = callbackFunc;
        workDone = 0;
        totalWork = numAssets - idx;
    };
    if (wrap) then
        self.searchRight.stopIdx = idx;
        self.searchRight.totalWork = self.searchRight.totalWork + idx;
    end
    local function workFunc(workToDo)
        local state = self.searchRight;
        state.workDone = state.workDone + workToDo;
        while (workToDo > 0) do
            local id = assetID[state.idx];
            if (self:MatchesFilter(id)) then
--Puts("Match: " .. state.idx .. "=>" .. string.format("0x%8.8X", id));
                table.insert(state.matches, state.idx);
                local done = (#state.matches >= state.matchesWanted);
                state.callbackFunc(state.matches, done);
                if (done) then
                    return 1; -- finished
                end
            end
            state.idx = state.idx + 1;
            if (state.idx > numAssets) then
                state.idx = 1;
            end
            if (state.idx == state.stopIdx) then
                -- Operation complete, but didn't find the requested number of matches.
                state.callbackFunc(state.matches, true);
                return 1; -- finished
            end
            workToDo = workToDo - 1;
        end
        return (state.workDone / state.totalWork);
    end
    self.rightProgressBar:Cancel();
    self.rightProgressBar:Start(workFunc, 10, 100);
end

function AssetBrowser:CancelSearches()
    self.leftProgressBar:Cancel();
    self.rightProgressBar:Cancel();
end

_G.recursion = 0;
-- See comment for SearchRight().
function AssetBrowser:_SearchLeft(idx, matchesWanted, wrap, callbackFunc)
    if ((idx == nil) or ((not wrap) and (idx == 1))) then
        -- Nothing to do.
        callbackFunc({}, true);
        return;
    end
    self.searchLeft = {
        idx = idx - 1;
        matchesWanted = matchesWanted;
        stopIdx = numAssets;
        matches = {};
        callbackFunc = callbackFunc;
        workDone = 0;
        totalWork = idx - 1;
    };
    if (wrap) then
        self.searchLeft.stopIdx = idx;
        self.searchLeft.totalWork = self.searchLeft.totalWork + 1 + (numAssets - idx);
    end
    local function workFunc(workToDo)
        local state = self.searchLeft;
        state.workDone = state.workDone + workToDo;
        while (workToDo > 0) do
            local id = assetID[state.idx];
            if (self:MatchesFilter(id)) then
                table.insert(state.matches, state.idx);
                local done = (#state.matches >= state.matchesWanted);
                state.callbackFunc(state.matches, done);
                if (done) then
                    return 1; -- finished
                end
            end
            state.idx = state.idx - 1;
            if (state.idx < 1) then
                state.idx = numAssets;
            end
            if (state.idx == state.stopIdx) then
                -- Operation complete, but didn't find the requested number of matches.
                state.callbackFunc(state.matches, true);
                return 1; -- finished
            end
            workToDo = workToDo - 1;
        end
        return (state.workDone / state.totalWork);
    end
    self.leftProgressBar:Cancel();
    self.leftProgressBar:Start(workFunc, 10, 100);
end

function AssetBrowser:Update()
    self:ValidateFilter();

    local scrollBar = self.assetIDScrollBar;
    if (scrollBar:GetValue() ~= scrollBar.lastKnownValue) then
        self:InvalidateImage();
        local idx = scrollBar:GetValue();
        scrollBar:SetValue(idx);
        if (idx > self.assetIDIndex) then
            self:_SearchRight(idx - 1, 1, false, function(matches, done)
                if (matches[1]) then
                    self.assetIDIndex = matches[1];
                    scrollBar:SetValue(self.assetIDIndex);
                    self:UpdateNeeded();
                end
            end);
        else
            self:_SearchLeft(idx + 1, 1, false, function(matches, done)
                if (matches[1]) then
                    self.assetIDIndex = matches[1];
                    scrollBar:SetValue(self.assetIDIndex);
                    self:UpdateNeeded();
                end
            end);
        end
    end

    local textBox = self.assetIDField;
    if (textBox:GetText() ~= textBox.lastKnownValue) then
        textBox.lastKnownValue = textBox:GetText();
        self:InvalidateImage();
        local id, idx = tonumber(textBox.lastKnownValue);
        if (id ~= nil) then
            idx = assetIDIndex[id];
        end
        if (idx) then
            self.assetIDIndex = idx;
            self:UpdateNeeded();
        end
    end

    if (self.updateNeeded) then
        local currentTime = Turbine.Engine.GetGameTime();
        if ((currentTime - self.filterChangeTime) >= .5) then
            self.imageOverlay:SetVisible(true);
            self:SelectAsset(self.assetIDIndex);
            self.updateNeeded = false;
        end
    end
end

function AssetBrowser:ValidateFilter()
    local paramNames = { "wantOld", "wantUd20x", "wantUd21x", "wantUd22x", "wantUd23x", "wantUd24x", "widthMin", "widthMax", "heightMin", "heightMax", "aspectMin", "aspectMax", "assetIDField" };
    local params = self.filterParams or {};
    local invalid = {};
    local changed = false;

    -- Get values from controls, check for changes, make sure they are valid
    for _, name in ipairs(paramNames) do
        local value = self[name]:GetValue();
        if (params[name] ~= value) then
            changed = true;
        end
        if (value == nil) then
            invalid[name] = true;
        end
        params[name] = value;
    end

    if (not changed) then
        -- Nothing to do.
        return;
    end

    -- Further checks for validity
    if ((params.widthMin and params.widthMax) and (params.widthMin > params.widthMax)) then
        invalid.widthMin = true;
        invalid.widthMax = true;
    end
    if ((params.heightMin and params.heightMax) and (params.heightMin > params.heightMax)) then
        invalid.heightMin = true;
        invalid.heightMax = true;
    end
    if ((params.aspectMin and params.aspectMax) and (params.aspectMin > params.aspectMax)) then
        invalid.aspectMin = true;
        invalid.aspectMax = true;
    end

    -- Invalid red; value values white
    local invalidate;
    for _, name in ipairs(paramNames) do
        if (invalid[name]) then
            self[name]:SetForeColor(Turbine.UI.Color.Red);
            invalidate = true;
        else
            self[name]:SetForeColor(Turbine.UI.Color.White);
        end
    end

    -- Update the image display area.
    self.filterParams = params;
    self.assetIDIndex = params.assetIDField;
    if (invalidate) then
        self:InvalidateImage();
    else
        self:UpdateNeeded();
    end
end

function AssetBrowser:ShowRenderMenu()
    local menu = Turbine.UI.ContextMenu();
    local items, subItems = menu:GetItems();

    item = Turbine.UI.MenuItem("BlendMode");
    items:Add(item);
    subItems = item:GetItems();

    for _, blendMode in ipairs({
        "AlphaBlend", "Color", "Grayscale", "Multiply",
        "None", "Normal", "Overlay", "Screen", "(nil)"
    }) do
        local value = Turbine.UI.BlendMode[blendMode];
        local checked = (self.imgBlendMode == value);
        local subItem = Turbine.UI.MenuItem(blendMode, true, checked);
        subItem.Click = function()
            self.imgBlendMode = value;
            self:UpdateImageDisplay();
        end
        subItems:Add(subItem);
    end
    
    item = Turbine.UI.MenuItem("BackColor");
    items:Add(item);
    subItems = item:GetItems();

    for _, color in ipairs({
        "White", "Black", "Red", "Orange", "Yellow",
        "Green", "Blue", "Indigo", "Violet", "(nil)"
    }) do
        local value = Turbine.UI.Color[color];
        local checked = (self.imgBackColor == value);
        local subItem = Turbine.UI.MenuItem(color, true, checked);
        subItem.Click = function()
            self.imgBackColor = value;
            self:UpdateImageDisplay();
        end
        subItems:Add(subItem);
    end

    item = Turbine.UI.MenuItem("BackColorBlendMode");
    items:Add(item);
    subItems = item:GetItems();
    
    for _, blendMode in ipairs({
        "AlphaBlend", "Color", "Grayscale", "Multiply",
        "None", "Normal", "Overlay", "Screen", "(nil)"
    }) do
        local value = Turbine.UI.BlendMode[blendMode];
        local checked = (self.imgBackColorBlendMode == value);
        local subItem = Turbine.UI.MenuItem(blendMode, true, checked);
        subItem.Click = function()
            self.imgBackColorBlendMode = value;
            self:UpdateImageDisplay();
        end
        subItems:Add(subItem);
    end

    item = Turbine.UI.MenuItem("Outline", true, self.imgOutline);
    items:Add(item);
    item.Click = function()
        self.imgOutline = not self.imgOutline;
        self:UpdateImageDisplay();
    end

    local x, y = self.renderLabel:PointToScreen(0, 0);
    menu:ShowMenuAt(x - 18, y + 18);
end
