ProgressBar = class(Turbine.UI.Control);

-- 'workFunc' is the work function, called each Update() cycle
--   prototype:
--     function workFunc(workToDo)
--       return progress;
--     end
--   where:
--     'workToDo' is an integer number of loop iterations to try
--     'progress' is a fraction between 0 and 1 indicating the amount of total progress so far
-- 'targetFrameRate' is the target frame rate.  Work will be increased/decreased as necessary to achieve this target.
-- 'initWorkToDo' is the 'workToDo' value to be passed in the first call to 'workFunc' -- an initial guess.

function ProgressBar:Constructor(workFunc, targetFrameRate, initWorkToDo)
    Turbine.UI.Control.Constructor(self);
    self:SetBackground(0x41007DF7);
    Turbine.UI.Control.SetHeight(self, 9);
    
    self.bar = Turbine.UI.Control();
    self.bar:SetParent(self);
    self.bar:SetBackground(0x41007DF5);
    self.bar:SetHeight(9);
    
    if (workFunc) then
        self:Start(workFunc, targetFrameRate, initWorkToDo);
    end
end

function ProgressBar:Start(workFunc, targetFrameRate, initWorkToDo)
    self.workFunc = workFunc;
    targetFrameRate = targetFrameRate or 10;
    self.targetTime = 1 / targetFrameRate;

    -- Do the first workFunc call immediately:
    self.prevWorkToDo = initWorkToDo or 1;
    self.prevGameTime = Turbine.Engine.GetGameTime();
    self:SetWantsUpdates(true); -- and continue at next Update() cycle, if necessary
    self:SetProgress(self.workFunc(self.prevWorkToDo));
end

function ProgressBar:Cancel()
    self:SetWantsUpdates(false);
end

function ProgressBar:Update()
    local gameTime = Turbine.Engine.GetGameTime();
    local elapsedTime = gameTime - self.prevGameTime;
    if (elapsedTime == 0) then
        return;
    end
    
    local workToDo = self.prevWorkToDo * (self.targetTime / elapsedTime);
    workToDo = math.max(1, math.floor(0.5 + workToDo));
    self:SetProgress(self.workFunc(workToDo));
    
    self.prevWorkToDo = workToDo;
    self.prevGameTime = gameTime;
end

function ProgressBar:SetProgress(progress)
    progress = math.max(0, math.min(1, progress));
    self.bar:SetWidth(math.floor(0.5 + self:GetWidth() * progress));
    if (progress == 1) then
        self:SetWantsUpdates(false);
    end
    self.progress = progress;
    DoCallbacks(self, "ProgressChanged", progress);
end

function ProgressBar:GetProgress()
    return self.progress;
end

function ProgressBar:SetHeight(height)
    self.bar:SetHeight(height);
    Turbine.UI.Control.SetHeight(self, height);
end

function ProgressBar:SetSize(width, height)
    self:SetWidth(width);
    self:SetHeight(height);
end

function ProgressBar:SetForeColor(color)
    self.bar:SetBackColor(color);
end

Thurallor = Thurallor or {};
Thurallor.UI = Thurallor.UI or {};
Thurallor.UI.ProgressBar = ProgressBar;
